#ifndef __BUF_H
#define __BUF_H

#include "list.h"
#include "pipe.h"

struct event;
typedef void (*CALLBACK)(struct buffer*,struct event*);

#ifndef SWIG

struct event
{
  struct link k;
  int i; /* when */
  CALLBACK cb;
};

typedef struct event*EVENT;

struct buffer
{
  struct link k;
  int id;

  /* ring buffer */
  char* mem; /* = malloc(size) */
  unsigned int size;

  int null; /* start reading here */
  int offset; /* start writing here */

  struct pipe *readers,*writers; /* ll */
/*  int in, out; */
/*  void (*send)(struct buffer*); */

#ifdef OBSOLETE
  struct list wev, rev; /* events pending */
  int i;
#endif
};

#endif

typedef struct buffer*BUFFER; 
typedef void (*UPDATE)(struct buffer*);

struct buffer*
buffer_new(unsigned int size); 
void 
buffer_reset(struct buffer*b);
int
buffer_invariant(struct buffer*b);
void
buffer_print(struct buffer*b );
int
buffer_id(struct buffer*b);


void
buffer_attach_reader(struct buffer*b,struct pipe*p);
void
buffer_attach_writer(struct buffer*b,struct pipe*p);
void
buffer_detach_reader(struct buffer*b,struct pipe*p);
void
buffer_detach_writer(struct buffer*b,struct pipe*p);

#ifdef OBSOLETE
void
write_event_init(struct event*e,struct buffer*b,int i,CALLBACK cb);
void
write_event_free(struct buffer*b, struct event*e);
#endif

int
buffer_send_read(struct buffer*b);
int
buffer_send_write(struct buffer*b);
int
buffer_send(struct buffer*b);

unsigned int
buffer_read_size(struct buffer*b, int i);
unsigned int
buffer_write_size(struct buffer*b, int i);
void*
buffer_peek_read(struct buffer*b,int i);
void*
buffer_peek_write(struct buffer*b,int i);

unsigned int
buffer_request_read(struct buffer*b, int i, int size);
unsigned int
buffer_request_write(struct buffer*b, int i, int size);

void 
buffer_free(struct buffer*b);


#endif

